@@ -15,14 +15,14 @@ using namespace ENC;
 #define PRIM_L 2
 #define PRIM_A 3
 
-constexpr quint32 SG2D = ISO8211::NAME("SG2D");
-constexpr quint32 SG3D = ISO8211::NAME("SG3D");
-constexpr quint32 FSPT = ISO8211::NAME("FSPT");
-constexpr quint32 VRPT = ISO8211::NAME("VRPT");
-constexpr quint32 ATTF = ISO8211::NAME("ATTF");
-constexpr quint32 VRID = ISO8211::NAME("VRID");
-constexpr quint32 FRID = ISO8211::NAME("FRID");
-constexpr quint32 DSPM = ISO8211::NAME("DSPM");
+constexpr quint32 SG2D = ISO8211::TAG("SG2D");
+constexpr quint32 SG3D = ISO8211::TAG("SG3D");
+constexpr quint32 FSPT = ISO8211::TAG("FSPT");
+constexpr quint32 VRPT = ISO8211::TAG("VRPT");
+constexpr quint32 ATTF = ISO8211::TAG("ATTF");
+constexpr quint32 VRID = ISO8211::TAG("VRID");
+constexpr quint32 FRID = ISO8211::TAG("FRID");
+constexpr quint32 DSPM = ISO8211::TAG("DSPM");
 
 static QMap<uint,uint> orderMapInit()
 {
@@ -85,6 +85,9 @@ static QMap<uint,uint> orderMapInit()
 	map.insert(TYPE(LNDELV), 44);
 	map.insert(TYPE(SMCFAC), 45);
 	map.insert(TYPE(BUISGL), 46);
+	map.insert(TYPE(ACHARE), 47);
+	map.insert(TYPE(I_ACHARE), 47);
+	map.insert(TYPE(DMPGRD), 48);
 
 	map.insert(TYPE(I_DISMAR), 0xFFFFFFFE);
 	map.insert(TYPE(SOUNDG), 0xFFFFFFFF);
@@ -151,9 +154,9 @@ static const ISO8211::Field *SGXD(const ISO8211::Record &r)
 {
 	const ISO8211::Field *f;
 
-	if ((f = ISO8211::field(r, SG2D)))
+	if ((f = r.field(SG2D)))
 		return f;
-	else if ((f = ISO8211::field(r, SG3D)))
+	else if ((f = r.field(SG3D)))
 		return f;
 	else
 		return 0;
@@ -188,17 +191,21 @@ static bool polygonPointCb(const MapData::Poly *polygon, void *context)
 
 	if (baseType == TSSLPT || baseType == RCTLPT || baseType == I_TRNBSN
 	  || baseType == BRIDGE || baseType == I_BRIDGE || baseType == BUAARE
-	  || baseType == LNDARE || baseType == LNDRGN
+	  || baseType == LNDARE || baseType == LNDRGN || baseType == I_BUNSTA
+	  || baseType == PILBOP || baseType == DMPGRD
 	  || type == SUBTYPE(ACHARE, 2) || type == SUBTYPE(I_ACHARE, 2)
 	  || type == SUBTYPE(ACHARE, 3) || type == SUBTYPE(I_ACHARE, 3)
 	  || type == SUBTYPE(ACHARE, 9) || type == SUBTYPE(I_ACHARE, 9)
-	  || type == SUBTYPE(I_BERTHS, 6)
+	  || type == SUBTYPE(I_ACHARE, 10) || type == SUBTYPE(I_ACHARE, 11)
+	  || type == SUBTYPE(I_ACHARE, 12) || type == SUBTYPE(I_BERTHS, 6)
 	  || type == SUBTYPE(RESARE, 1) || type == SUBTYPE(I_RESARE, 1)
 	  || type == SUBTYPE(RESARE, 2) || type == SUBTYPE(I_RESARE, 2)
+	  || type == SUBTYPE(RESARE, 3) || type == SUBTYPE(I_RESARE, 3)
 	  || type == SUBTYPE(RESARE, 4) || type == SUBTYPE(I_RESARE, 4)
 	  || type == SUBTYPE(RESARE, 5) || type == SUBTYPE(I_RESARE, 5)
 	  || type == SUBTYPE(RESARE, 6) || type == SUBTYPE(I_RESARE, 6)
 	  || type == SUBTYPE(RESARE, 7) || type == SUBTYPE(I_RESARE, 7)
+	  || type == SUBTYPE(RESARE, 8) || type == SUBTYPE(I_RESARE, 8)
 	  || type == SUBTYPE(RESARE, 9) || type == SUBTYPE(I_RESARE, 9)
 	  || type == SUBTYPE(RESARE, 12) || type == SUBTYPE(I_RESARE, 12)
 	  || type == SUBTYPE(RESARE, 14) || type == SUBTYPE(I_RESARE, 14)
@@ -208,7 +215,7 @@ static bool polygonPointCb(const MapData::Poly *polygon, void *context)
 	  || type == SUBTYPE(RESARE, 25) || type == SUBTYPE(I_RESARE, 25)
 	  || type == SUBTYPE(RESARE, 26) || type == SUBTYPE(I_RESARE, 26))
 		points->append(MapData::Point(baseType, polygon->bounds().center(),
-		  polygon->attributes(), polygon->HUNI(), true));
+		  polygon->attributes(), polygon->huni(), true));
 
 	return true;
 }
@@ -318,6 +325,8 @@ static uint restrictionCategory(uint type, const MapData::Attributes &attr)
 
 		if (restrn == 1)
 			return 2;
+		else if (restrn == 3)
+			return 3;
 		else if (restrn == 7)
 			return 17;
 		else
@@ -326,6 +335,20 @@ static uint restrictionCategory(uint type, const MapData::Attributes &attr)
 		return catrea;
 }
 
+static uint color(const QList<QByteArray> &list)
+{
+	uint c = 0;
+
+	for (int i = 0, j = 0; i < list.size() && j < 4; i++) {
+		if (!(i && list.at(i) == list.at(i-1))) {
+			c |= list.at(i).toUInt() << (j * 4);
+			j++;
+		}
+	}
+
+	return c;
+}
+
 MapData::Point::Point(uint type, const Coordinates &c, const QString &label)
   : _type(SUBTYPE(type, 0)), _pos(c), _label(label), _polygon(false)
 {
@@ -336,6 +359,7 @@ MapData::Point::Point(uint type, const Coordinates &c, const Attributes &attr,
   uint HUNI, bool polygon) : _pos(c), _attr(attr), _polygon(polygon)
 {
 	uint subtype = 0;
+	bool ok;
 
 	if (type == HRBFAC)
 		subtype = CATHAF;
@@ -375,16 +399,26 @@ MapData::Point::Point(uint type, const Coordinates &c, const Attributes &attr,
 		subtype = CATACH;
 	else if (type == I_ACHARE)
 		subtype = I_CATACH;
-	else if (type == MARKUL)
+	else if (type == MARCUL)
 		subtype = CATMFA;
 	else if (type == I_BUNSTA)
 		subtype = I_CATBUN;
+	else if (type == BOYCAR || type == BOYINB || type == BOYISD
+	  || type == BOYLAT || type == I_BOYLAT || type == BOYSAW || type == BOYSPP
+	  || type == BCNCAR || type == BCNISD || type == BCNLAT || type == I_BCNLAT
+	  || type == BCNSAW || type == BCNSPP)
+		subtype = COLOUR;
 
 	QList<QByteArray> list(_attr.value(subtype).split(','));
-	std::sort(list.begin(), list.end());
-	_type = (type == RESARE || type == I_RESARE)
-	  ? SUBTYPE(type, restrictionCategory(type, _attr))
-	  : SUBTYPE(type, list.first().toUInt());
+	if (type == RESARE || type == I_RESARE)
+		_type = SUBTYPE(type, restrictionCategory(type, _attr));
+	else if (subtype == COLOUR)
+		_type = SUBTYPE(type, color(list));
+	else {
+		std::sort(list.begin(), list.end());
+		_type = SUBTYPE(type, list.first().toUInt());
+	}
+
 	_id = ((quint64)order(_type))<<32 | (uint)qHash(c);
 	_label = QString::fromLatin1(_attr.value(OBJNAM));
 
@@ -417,16 +451,22 @@ MapData::Point::Point(uint type, const Coordinates &c, const Attributes &attr,
 			_label += "\n(" + QString::fromLatin1(_attr.value(ELEVAT))
 			  + "\xE2\x80\x89m)";
 	} else if (type == BRIDGE || type == I_BRIDGE) {
-		double clr = _attr.value(VERCLR).toDouble();
-		if (clr > 0) {
+		double clr = _attr.value(VERCLR).toDouble(&ok);
+		if (ok && clr > 0)
 			_label = QString::fromUtf8("\xE2\x86\x95") + UNIT_SPACE
 			  + QString::number(clr) + UNIT_SPACE + hUnits(HUNI);
-		}
+	} else if (type == OBSTRN || type == WRECKS || type == UWTROC) {
+		double depth = _attr.value(VALSOU).toDouble(&ok);
+		if (ok && _label.isEmpty())
+			_label = QString::number(depth);
+	} else if (_type == SUBTYPE(RESARE, 8)) {
+		if (_label.isEmpty())
+			_label = "Degaussing Range";
 	}
 }
 
 MapData::Poly::Poly(uint type, const Polygon &path, const Attributes &attr,
-  uint HUNI) : _path(path), _attr(attr), _HUNI(HUNI)
+  uint HUNI) : _path(path), _attr(attr), _huni(HUNI)
 {
 	uint subtype = 0;
 
@@ -436,12 +476,20 @@ MapData::Poly::Poly(uint type, const Polygon &path, const Attributes &attr,
 		subtype = I_CATACH;
 	else if (type == HRBFAC)
 		subtype = CATHAF;
-	else if (type == MARKUL)
+	else if (type == MARCUL)
 		subtype = CATMFA;
 	else if (type == I_BERTHS)
 		subtype = I_CATBRT;
 	else if (type == M_COVR)
 		subtype = CATCOV;
+	else if (type == DMPGRD)
+		subtype = CATDPG;
+	else if (type == SLCONS)
+		subtype = CATSLC;
+	else if (type == I_SLCONS)
+		subtype = I_CATSLC;
+	else if (type == ADMARE)
+		subtype = JRSDTN;
 
 	switch (type) {
 		case DEPARE:
@@ -477,6 +525,8 @@ MapData::Line::Line(uint type, const QVector<Coordinates> &path,
 		_label = QString::fromLatin1(_attr.value(VALDCO));
 	else if (type == LNDELV)
 		_label = QString::fromLatin1(_attr.value(ELEVAT));
+	else if (type == RECTRC)
+		_label = QString::fromLatin1(_attr.value(ORIENT));
 	else
 		_label = QString::fromLatin1(_attr.value(OBJNAM));
 }
@@ -495,7 +545,7 @@ QVector<MapData::Sounding> MapData::soundings(const ISO8211::Record &r,
   uint comf, uint somf)
 {
 	QVector<Sounding> s;
-	const ISO8211::Field *f = ISO8211::field(r, SG3D);
+	const ISO8211::Field *f = r.field(SG3D);
 	if (!f)
 		return QVector<Sounding>();
 
@@ -517,7 +567,7 @@ QVector<MapData::Sounding> MapData::soundingGeometry(const ISO8211::Record &r,
 	quint32 id;
 	RecordMapIterator it;
 
-	const ISO8211::Field *fspt = ISO8211::field(r, FSPT);
+	const ISO8211::Field *fspt = r.field(FSPT);
 	if (!fspt || fspt->data().at(0).size() != 4)
 		return QVector<Sounding>();
 
@@ -545,7 +595,7 @@ Coordinates MapData::pointGeometry(const ISO8211::Record &r,
 	quint32 id;
 	RecordMapIterator it;
 
-	const ISO8211::Field *fspt = ISO8211::field(r, FSPT);
+	const ISO8211::Field *fspt = r.field(FSPT);
 	if (!fspt || fspt->data().at(0).size() != 4)
 		return Coordinates();
 
@@ -575,7 +625,7 @@ QVector<Coordinates> MapData::lineGeometry(const ISO8211::Record &r,
 	quint8 type;
 	quint32 id;
 
-	const ISO8211::Field *fspt = ISO8211::field(r, FSPT);
+	const ISO8211::Field *fspt = r.field(FSPT);
 	if (!fspt || fspt->data().at(0).size() != 4)
 		return QVector<Coordinates>();
 
@@ -588,7 +638,7 @@ QVector<Coordinates> MapData::lineGeometry(const ISO8211::Record &r,
 		if (it == ve.constEnd())
 			return QVector<Coordinates>();
 		const ISO8211::Record &frid = it.value();
-		const ISO8211::Field *vrpt = ISO8211::field(frid, VRPT);
+		const ISO8211::Field *vrpt = frid.field(VRPT);
 		if (!vrpt || vrpt->data().size() != 2)
 			return QVector<Coordinates>();
 
@@ -641,7 +691,7 @@ Polygon MapData::polyGeometry(const ISO8211::Record &r, const RecordMap &vc,
 	quint8 type;
 	quint32 id;
 
-	const ISO8211::Field *fspt = ISO8211::field(r, FSPT);
+	const ISO8211::Field *fspt = r.field(FSPT);
 	if (!fspt || fspt->data().at(0).size() != 4)
 		return Polygon();
 
@@ -660,7 +710,7 @@ Polygon MapData::polyGeometry(const ISO8211::Record &r, const RecordMap &vc,
 		if (it == ve.constEnd())
 			return Polygon();
 		const ISO8211::Record &frid = it.value();
-		const ISO8211::Field *vrpt = ISO8211::field(frid, VRPT);
+		const ISO8211::Field *vrpt = frid.field(VRPT);
 		if (!vrpt || vrpt->data().size() != 2)
 			return Polygon();
 
@@ -723,7 +773,7 @@ MapData::Attributes MapData::attributes(const ISO8211::Record &r)
 {
 	Attributes attr;
 
-	const ISO8211::Field *attf = ISO8211::field(r, ATTF);
+	const ISO8211::Field *attf = r.field(ATTF);
 	if (!(attf && attf->data().at(0).size() == 2))
 		return attr;
 
@@ -829,9 +879,15 @@ MapData::MapData(const QString &path)
 
 	if (!ddf.readDDR())
 		return;
-	while (ddf.readRecord(record))
+	while (!ddf.atEnd()) {
+		if (!ddf.readRecord(record)) {
+			qWarning("%s: %s", qUtf8Printable(path),
+			  qUtf8Printable(ddf.errorString()));
+			return;
+		}
 		if (!processRecord(record, fe, vi, vc, ve, comf, somf, huni))
-			qWarning("Invalid S-57 record");
+			qWarning("%s: Invalid S-57 record", qUtf8Printable(path));
+	}
 
 	for (int i = 0; i < fe.size(); i++) {
 		const ISO8211::Record &r = fe.at(i);
